#ifndef _ADT_H
#define _ADT_H

/*{{{  includes */

#include <aterm1.h>
#include "ADT_dict.h"

/*}}}  */

/*{{{  prologue */

typedef ATerm ADTSort;
typedef ATerm ADTAlternative;
typedef ATerm ADTPattern;

/*}}}  */
/*{{{  typedefs */

typedef struct _ADTEntries *ADTEntries;
typedef struct _ADTEntry *ADTEntry;

/*}}}  */

void ADTinitADTApi(void);

/*{{{  term conversion functions */

ADTEntries ADTEntriesFromTerm(ATerm t);
ATerm ADTEntriesToTerm(ADTEntries arg);
ADTEntry ADTEntryFromTerm(ATerm t);
ATerm ADTEntryToTerm(ADTEntry arg);

/*}}}  */
/*{{{  constructors */

ADTEntries ADTmakeEntriesEmpty();
ADTEntries ADTmakeEntriesList(ADTEntry head, ADTEntries tail);
ADTEntry ADTmakeEntryDefault(ADTSort sort, ADTAlternative alt, ADTPattern pat);

/*}}}  */
/*{{{  equality functions */

ATbool ADTisEqualEntries(ADTEntries arg0, ADTEntries arg1);
ATbool ADTisEqualEntry(ADTEntry arg0, ADTEntry arg1);

/*}}}  */
/*{{{  ADTEntries accessors */

ATbool ADTisValidEntries(ADTEntries arg);
inline ATbool ADTisEntriesEmpty(ADTEntries arg);
inline ATbool ADTisEntriesList(ADTEntries arg);
ATbool ADThasEntriesHead(ADTEntries arg);
ADTEntry ADTgetEntriesHead(ADTEntries arg);
ADTEntries ADTsetEntriesHead(ADTEntries arg, ADTEntry head);
ATbool ADThasEntriesTail(ADTEntries arg);
ADTEntries ADTgetEntriesTail(ADTEntries arg);
ADTEntries ADTsetEntriesTail(ADTEntries arg, ADTEntries tail);

/*}}}  */
/*{{{  ADTEntry accessors */

ATbool ADTisValidEntry(ADTEntry arg);
inline ATbool ADTisEntryDefault(ADTEntry arg);
ATbool ADThasEntrySort(ADTEntry arg);
ADTSort ADTgetEntrySort(ADTEntry arg);
ADTEntry ADTsetEntrySort(ADTEntry arg, ADTSort sort);
ATbool ADThasEntryAlt(ADTEntry arg);
ADTAlternative ADTgetEntryAlt(ADTEntry arg);
ADTEntry ADTsetEntryAlt(ADTEntry arg, ADTAlternative alt);
ATbool ADThasEntryPat(ADTEntry arg);
ADTPattern ADTgetEntryPat(ADTEntry arg);
ADTEntry ADTsetEntryPat(ADTEntry arg, ADTPattern pat);

/*}}}  */
/*{{{  sort visitors */

ADTEntries ADTvisitEntries(ADTEntries arg, ADTEntry (*acceptHead)(ADTEntry));
ADTEntry ADTvisitEntry(ADTEntry arg, ADTSort (*acceptSort)(ADTSort), ADTAlternative (*acceptAlt)(ADTAlternative), ADTPattern (*acceptPat)(ADTPattern));

/*}}}  */

#endif /* _ADT_H */
